'''
Relativictic Difference Scheme Particle-in-Cell code (REDPIC) constants file.

'''

import periodictable
import numpy as np
from collections import namedtuple
from scipy import constants

__all__ = [ 'speed_of_light',
            'c',
            'epsilon_0',
            'ep_0',
            'mu_0',
            'h',
            'hbar',
            'elementary_charge',
            'e',
            'electron_mass',
            'm_e',
            'electron_mass_energy',
            'mc',
            'electron_radius',
            'r_0',
            'proton_mass',
            'm_p',
            'neutron_mass',
            'm_n',
            'atomic_constant_mass',
            'm_u',
            'u',
            'electron',
            'positron',
            'proton',
            'antiproton',
            'neutron',
            'antineutron',
            'Element',
            'Particle'
            ]

# Constants
c = speed_of_light = constants.c

ep_0 = epsilon_0 = constants.epsilon_0
mu_0 = constants.mu_0
h = constants.h
hbar = constants.hbar

e = elementary_charge = constants.e

m_e = electron_mass = constants.m_e
mc = electron_mass_energy = constants.physical_constants[ 'electron mass energy equivalent in MeV' ][0]
r_0 = electron_radius = constants.physical_constants[ 'classical electron radius' ][0]
m_p = proton_mass = constants.m_p
m_n = neutron_mass = constants.m_n
u = m_u = atomic_constant_mass = constants.physical_constants[ 'atomic mass constant' ][0]

Element = namedtuple('Element', [ 'name', 'symbol', 'mass', 'charge' ])

electron = Element(name='electron', symbol='e', mass=m_e, charge=-e)
positron = Element(name='positron', symbol='e+', mass=m_e, charge=e)
proton = Element(name='proton', symbol='p', mass=m_p, charge=e)
antiproton = Element(name='antiproton', symbol='p-', mass=m_p, charge=-e)
neutron = Element(name='neutron', symbol='n', mass=m_n, charge=0)
antineutron = Element(name='antineutron', symbol='n', mass=m_n, charge=0)
Particle = Element

# Get mass of each element from periodictable
elements = periodictable.core.default_table()
__all__  += periodictable.core.define_elements(elements, globals())



'''
Relativictic Difference Scheme Particle-in-Cell code (REDPIC) init file.

'''

from .constants import *
from .beam import *
from .accelerator import *
from .solver import *

__version__ = '0.1.0'
__doc__ = '''Relativictic Difference Scheme Particle-in-Cell code (REDPIC)'''

__all__ = constants.__all__ + beam.__all__ + solver.__all__ + accelerator.__all__



'''
Relativictic Difference Scheme Particle-in-Cell code (REDPIC) solver file.

'''

import numpy as np
import pandas as pd
from scipy import misc
from .constants import *
from .beam import *

__all__ = [ 'Simulation',
            'Sim'
             ]

class Simulation:
    ''' Simulation of the envelope beam in the accelerator

    '''
    def __init__(self, beam, accelerator):
        self.beam = beam
        self.acc = accelerator

    def track(self):
        ''' Tracking!

        '''

        # Constants
        m = self.beam.type.mass
        q = self.beam.type.charge
        dz = self.acc.dz
        dt = dz/c
        t_max = (self.acc.z_stop-self.acc.z_start)/c
        Y = self.beam.da
        P0 = m*c*c / (e*1e6)
        E0 = m*c / (e*dt*1e6)
        B0 = m / (e*dt)

        # RED
        for t in np.arange(0, t_max, 2*dt):
            Y[0], Y[1], Y[2] = Y[0]/dz, Y[1]/dz, Y[2]/dz
            Y[3], Y[4], Y[5] = Y[3]/P0, Y[4]/P0, Y[5]/P0


            offset_x = self.acc.Dx(Y[2]*dz)
            offset_xp = self.acc.Dxp(Y[2]*dz)
            offset_y = self.acc.Dy(Y[2]*dz)
            offset_yp = self.acc.Dyp(Y[2]*dz)
            x_corr = Y[0]*dz - offset_x
            y_corr = Y[1]*dz - offset_y
            r_corr = np.sqrt(x_corr*x_corr + y_corr*y_corr)
            Ez = self.acc.Ez(Y[2]*dz)
            dEzdz = self.acc.dEzdz(Y[2]*dz)
            d2Ezdz2 = misc.derivative(self.acc.dEzdz, Y[2]*dz, dx=dz, n=1)
            Ez = (Ez - d2Ezdz2*r_corr*r_corr/4 - d2Ezdz2*r_corr*r_corr/4)/E0            # row remainder
            Ex = (- dEzdz*x_corr/2 - dEzdz*x_corr/2 + Ez*offset_xp)/E0          # row remainder
            Ey = (- dEzdz*y_corr/2 - dEzdz*y_corr/2 + Ez*offset_yp)/E0          # row remainder


            Y[3], Y[4], Y[5] = Y[3] - 2*Ex, Y[4] - 2*Ey, Y[5] - 2*Ez
            gamma = np.sqrt(1 + Y[3]*Y[3] + Y[4]*Y[4] + Y[5]*Y[5])
            vx, vy, vz = Y[3]/gamma, Y[4]/gamma, Y[5]/gamma
            Y[0], Y[1], Y[2] = Y[0] + vx, Y[1] + vy, Y[2] + vz


            offset_x = self.acc.Dx(Y[2]*dz)
            offset_xp = self.acc.Dxp(Y[2]*dz)
            offset_y = self.acc.Dy(Y[2]*dz)
            offset_yp = self.acc.Dyp(Y[2]*dz)
            x_corr = Y[0]*dz - offset_x
            y_corr = Y[1]*dz - offset_y
            r_corr = np.sqrt(x_corr*x_corr + y_corr*y_corr)
            Bx = self.acc.Bx(Y[2]*dz)
            By = self.acc.By(Y[2]*dz)
            Bz = self.acc.Bz(Y[2]*dz)
            dBzdz = self.acc.dBzdz(Y[2]*dz)
            d2Bzdz2 = misc.derivative(self.acc.dBzdz, Y[2]*dz, dx=dz, n=1)
            Gz = self.acc.Gz(Y[2]*dz)
            Bz = (Bz - d2Bzdz2*r_corr*r_corr/4 - d2Bzdz2*r_corr*r_corr/4)/B0/gamma              # row remainder
            Bx = (Bx + Gz*y_corr - dBzdz*x_corr/2 - dBzdz*x_corr/2 + Bz*offset_xp)/B0/gamma     # row remainder
            By = (By + Gz*x_corr - dBzdz*y_corr/2 - dBzdz*y_corr/2 + Bz*offset_yp)/B0/gamma     # row remainder


            b2 = 1 + Bx*Bx + By*By + Bz*Bz
            b1 = 2 - b2
            b3 = 2 * (vx*Bx + vy*By + vz*Bz)
            fx = 2 * (vy*Bz - vz*By)
            fy = 2 * (vz*Bx - vx*Bz)
            fz = 2 * (vx*By - vy*Bx)
            vx = (vx*b1 + fx + Bx*b3)/b2
            vy = (vy*b1 + fy + By*b3)/b2
            vz = (vz*b1 + fz + Bz*b3)/b2
            Y[0], Y[1], Y[2] = Y[0] + vx, Y[1] + vy, Y[2] + vz
            Y[3], Y[4], Y[5] = vx*gamma, vy*gamma, vz*gamma

            Y[0], Y[1], Y[2] = Y[0]*dz, Y[1]*dz, Y[2]*dz
            Y[3], Y[4], Y[5] = Y[3]*P0, Y[4]*P0, Y[5]*P0
            Bx, By, Bz = Bx*gamma*B0, By*gamma*B0, Bz*gamma*B0
            Ex, Ey, Ez = Ex*E0, Ey*E0, Ez*E0

            X = np.row_stack((Y[0], Y[1], Y[2], Y[3], Y[4], Y[5], Bx, By, Bz, Ex, Ey, Ez))
            Xt = np.transpose(X)
            df = pd.DataFrame(Xt, columns=['x','y','z','px','py','pz',
                             'Bx', 'By', 'Bz', 'Ex', 'Ey', 'Ez' ])
            progress = t/t_max*100
            meters = self.acc.z_start+t*c
            print( '\rz = %.2f m (%.1f %%) ' % (meters, progress), end='')
            if progress%3 == 0 or progress%4 == 0:
                df.to_csv(self.beam.type.symbol + 'Beam.'+ '%04.0f' % (meters*100) +'.csv')

Sim = Simulation



'''
Relativictic Difference Scheme Particle-in-Cell code (REDPIC) beam file.

'''

import numpy as np
import pandas as pd
from collections import namedtuple
from .constants import *

__all__ = [ 'Distribution',
            'Beam' ]

Distribution = namedtuple('Distribution', [ 'name', 'x', 'y', 'z', 'px', 'py', 'pz' ])

class Beam:
    '''Creates a beam of selected type particles

    '''

    def __init__(self, type: Element) -> None:
        self.type = type        # particles type
        self.n = 0.0            # quantity
        self.da = np.array      # data array
        self.df = pd.DataFrame  # data frame

    def generate(self, distribution: Distribution, *, n: float,
                 x_off: float=0.0, y_off: float=0.0, z_off: float=0.0, sig_pz: float=0.01) -> None:
        '''Beam generator

        This function generates a beam with a given distribution and initial beam displacement.
        '''
        assert n > 0, 'The number of particles (n) must be a positive number!'
        assert type(distribution.name) == str, 'Distribution name must be a string!'
        condition = ((distribution.name == 'KV' or distribution.name == 'Uniform') or
                    (distribution.name == 'Gauss' or distribution.name == 'GA'))
        assert condition, 'Сheck distribution name!'

        self.n = n

        if distribution.name == 'Uniform' or distribution.name == 'KV':
            phi = np.random.uniform(0, 2*np.pi, int(self.n))
            theta = np.random.uniform(0, np.pi, int(self.n))
            x = distribution.x * np.sin(theta) * np.cos(phi)
            y = distribution.y * np.sin(theta) * np.sin(phi)
            z = distribution.z * np.cos(theta)

        if distribution.name == 'Gauss' or distribution.name == 'GA':
            x = np.random.normal(x_off, distribution.x, int(self.n))
            y = np.random.normal(y_off, distribution.y, int(self.n))
            z = np.random.normal(z_off, distribution.z, int(self.n))

        px = np.random.normal(0, distribution.px, int(self.n))
        py = np.random.normal(0, distribution.py, int(self.n))
        pz = np.random.normal(distribution.pz, distribution.pz * sig_pz, int(self.n))

        self.da = np.row_stack((x, y, z, px, py, pz))
        self.df = pd.DataFrame(np.transpose(self.da), columns=['x','y','z','px','py','pz'])
        self.df.to_csv(self.type.symbol + 'Beam.csv')

    def upload(self, Y: np.array) -> None:
        ''' Particle loading

        Y = np.array[[ ... ] # x[m]
                     [ ... ] # y[m]
                     [ ... ] # z[m]
                     [ ... ] # px[MeV/c]
                     [ ... ] # py[MeV/c]
                     [ ... ] # pz[MeV/c]
                     ]
        '''
        self.da = Y
        self.df = pd.DataFrame(np.transpose(self.da), columns=['x','y','z','px','py','pz'])

    def __str__(self):
        return str(self.df)

    def _ipython_display_(self):
        print(str(self.df))
